import openpyxl as xl
from ..models.database.library import TableControl


class StyleControl():

    # 行ヘッダーは２列
    NUMBER_OF_COLUMNS_OF_ROW_HEADER = 2

    # ツリー区とプロパティ区を分けるセパレーターは１列
    SEPARATOR = 1

    # １つのノードは３列
    ONE_NODE_COLUMNS = 3


    @staticmethod
    def get_target_column_th(source_table, column_name):
        """書出し先ワークブックでの列番号を返します
        
        例： no, node0, node1, node2
        例： node0, node1, node2
        例： node0, edge1, node1, node2

        no 列、edge{数}列はオプションです

        FIXME: `node0, edge1, foo1, bar1, node1` のように予期しない列が混ざっているケースは？
        """

        # ルートノード
        #column_th_of_source_node_0 = TableControl.get_column_location_by_column_name(df=source_table.df, column_name='node0') + 1

        # 最終ノード
        column_th_of_source_last_node = source_table.analyzer.get_column_th_of_last_node()

        specified_column_location_of_source = TableControl.get_column_location_by_column_name(df=source_table.df, column_name=column_name)

        # 列名から、ノード、エッジ、余り列を見分ける
        # ノード
        result = TableControl.pattern_of_column_name_of_node.match(column_name)
        if result:
            node_th = int(result.group(1))
            return StyleControl.NUMBER_OF_COLUMNS_OF_ROW_HEADER + 1 + node_th * StyleControl.ONE_NODE_COLUMNS

        # エッジ
        result = TableControl.pattern_of_column_name_of_edge.match(column_name)
        if result:
            edge_th = int(result.group(1))
            # FIXME node の前に edge 列があるという前提で大丈夫か？
            return StyleControl.NUMBER_OF_COLUMNS_OF_ROW_HEADER + 1 + edge_th * StyleControl.ONE_NODE_COLUMNS - 1

        # それ以外

        # 書出し先のツリー区の最後
        last_column_th_of_wb = StyleControl.NUMBER_OF_COLUMNS_OF_ROW_HEADER + source_table.analyzer.end_th_of_node * StyleControl.ONE_NODE_COLUMNS

        # プロパティ区
        column_th_of_source_in_property_ward = specified_column_location_of_source - column_th_of_source_last_node -1     # 1 以上になる
        return last_column_th_of_wb + StyleControl.SEPARATOR + column_th_of_source_in_property_ward             # 空列を１つ挟む


    @staticmethod
    def get_number_of_character_of_column(df, column_name):
        """指定の列の文字数を取得"""

        series = df[column_name]
        #print(f"{type(series)=}")


        # 📖 [pandasのSeriesの要素の中で最大の桁数・文字数を取得する](https://qiita.com/Maron_T/items/105966b489c110b90ebe)
        if series.dtype == 'int64':
            return int(series.abs().apply("log10").max()) + 1

        # seriesが浮動小数点型なら小数点以下の桁数を返す
        if series.dtype == 'float64':
            return series.abs().astype(str).str.len().max()-1

        # seriesが文字列型なら文字数を返す
        if series.dtype == 'object':
            return series.str.len().max()

        raise ValueError(f'unsupported {series.dtype=}')




import gc
import os
import datetime
import openpyxl as xl
from .models.tree_structure import Forest
from .settings import Settings
from .worksheet_handle import WorksheetHandle


def planting():
    """森を作成します"""
    return Forest()


def prepare_workbook(target, mode, settings={}, debug_write=False):
    """ワークブックを用意します
    
    Returns
    -------
    target : str
        ワークブック（.xlsx）へのファイルパス
    mode : str
        既存のワークブックが有ったときの挙動。 'w' は新規作成して置換え、 'a' は追記
    settings : dict
        各種設定    
    debug_write : bool
        デバッグライト
    """

    wb = None

    # 既存のファイルが有ったときの挙動
    if os.path.isfile(target):
        # 既存のファイルへ追記
        if mode == 'a':
            if debug_write:
                print(f"[{datetime.datetime.now()}] `{target}` file exists, read.")

            # ワークブックを開く
            wb = xl.load_workbook(filename=target)
        
        elif mode == 'w':
            pass
        
        else:
            raise ValueError(f"unsupported {mode=}")

                
    # ワークブックを新規生成
    if wb is None:
        if debug_write:
            print(f"[{datetime.datetime.now()}] `{target}` file not exists, create.")

        wb = xl.Workbook()


    return WorkbookHandle(target=target, mode=mode, wb=wb, settings=settings, debug_write=debug_write)


class XltreeInSrc():
    """テストで使う仕組み

    テストでは

        import xltree as tr

    のように書けないので、
    テストでは以下のように書く

        from src.xltree import xltree_in_src as tr
    
    """


    @staticmethod
    def planting():
        """グローバル関数の planting() を呼び出す
        """
        global planting

        return planting()


    @staticmethod
    def prepare_workbook(target, mode, settings={}, debug_write=False):
        """グローバル関数の prepare_workbook() を呼び出す
        
        Returns
        -------
        target : str
            ワークブック（.xlsx）へのファイルパス
        mode : str
            既存のワークブックが有ったときの挙動。 'w' は新規作成して置換え、 'a' は追記
        settings : dict
            各種設定    
        debug_write : bool
            デバッグライト
        """
        global prepare_workbook

        return prepare_workbook(target=target, mode=mode, settings=settings, debug_write=debug_write)


xltree_in_src = XltreeInSrc()


class WorkbookHandle():
    """ワークブックへの対応"""


    def __init__(self, target, mode, wb, settings={}, debug_write=False):
        """初期化

        Parameters
        ----------
        target : str
            ワークブック（.xlsx）へのファイルパス
        mode : str
            既存のワークブックが有ったときの挙動。 'w' は新規作成して置換え、 'a' は追記
        settings : dict
            各種設定
        """
        self._wb_file_path = target
        self._mode = mode
        self._wb = wb
        self._settings_obj = Settings(dictionary=settings)
        self._debug_write = debug_write
        self._ws = None


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        del self._wb_file_path
        del self._mode
        del self._wb
        del self._settings_obj
        del self._debug_write
        del self._ws

        # メモリ解放
        gc.collect()


    @property
    def workbook_file_path(self):
        return self._wb_file_path


    def prepare_worksheet(self, target, based_on, debug_write=False):
        """ワークシートの用意"""

        # ワークシートの準備
        if not self.exists_sheet(target):
            if self._debug_write:
                print(f"[{datetime.datetime.now()}] create `{target}` sheet...")

            self._wb.create_sheet(target)


        ws = self._wb[target]


        # ワークシート制御の生成
        return WorksheetHandle.instantiate(target=target, based_on=based_on, ws=ws, settings_obj=self._settings_obj, debug_write=debug_write)


    def render_worksheet(self, target, based_on, debug_write=False):
        """ワークシートへ木構造図を描画

        Parameters
        ----------
        target : str
            シート名
        based_on : str
            CSVファイルパス
        debug_write : bool
            デバッグライト
        """

        # ワークシート制御の生成
        wsh = WorksheetHandle.instantiate(target=target, based_on=based_on, ws=self._ws, settings_obj=self._settings_obj, debug_write=debug_write)

        # 木の描画
        wsh.render_tree()


    def remove_worksheet(self, target):
        """存在すれば、指定のワークシートの削除。存在しなければ無視

        Parameters
        ----------
        target : str
            シート名
        """

        if self.exists_sheet(target=target):
            if self._debug_write:
                print(f"[{datetime.datetime.now()}] remove `{target}` sheet...")

            self._wb.remove(self._wb[target])


    def save_workbook(self):
        """ワークブックの保存"""

        if self._debug_write:
            print(f"[{datetime.datetime.now()}] save `{self._wb_file_path}` file...")

        # ワークブックの保存            
        self._wb.save(self._wb_file_path)


    def exists_sheet(self, target):
        """シートの存在確認
        
        Parameters
        ----------
        target : str
            シート名
        """
        return target in self._wb.sheetnames


    def get_worksheet(self, sheet_name):
        """ワークシートの取得"""
        return self._wb[sheet_name]



import re


class TableControl():


    _pattern_of_column_name_of_node = re.compile(r'node(\d+)')
    _pattern_of_column_name_of_edge = re.compile(r'edge(\d+)')


    @classmethod
    @property
    def pattern_of_column_name_of_node(clazz):
        return clazz._pattern_of_column_name_of_node


    @classmethod
    @property
    def pattern_of_column_name_of_edge(clazz):
        return clazz._pattern_of_column_name_of_edge


    @staticmethod
    def get_column_location_by_column_name(df, column_name):
        return df.columns.get_loc(column_name)


    @staticmethod
    def sort_out_column_names_n_o_node_edge_others(df):
        """列名を edge, node, それ以外の３つに分けます。
        edge と node は最後の要素の数字の +1 を返します。要素がなければ 0 を返します。
        'no' 列は、有るケースと無いケースがあります
        """
        is_n_o_column_exists = False
        edge_th_set = set()
        node_th_set = set()
        others_name_list = []

        # 'no' はインデックスなので、列には含まれない
        for column_name in df.columns.values:
            if column_name == 'no':
                is_n_o_column_exists = True
                continue
            
            result = TableControl.pattern_of_column_name_of_edge.match(column_name)
            if result:
                edge_th_set.add(int(result.group(1)))
                continue

            result = TableControl.pattern_of_column_name_of_node.match(column_name)
            if result:
                node_th_set.add(int(result.group(1)))
                continue
            
            others_name_list.append(column_name)
        
        # node 列は 0 から連番で続いているところまで有効にします
        end_th_of_node = 0
        for i in range(0, len(node_th_set)):
            if i not in node_th_set:
                break

            node_th_set.remove(i)
            end_th_of_node = i + 1

        # 残っている node 列は others リストに入れます
        for node_th in node_th_set:
            others_name_list.append(f'node{node_th}')

        # edge 列は 1 から ［end_th_of_node の手前］まで連番で続いているところだけ有効にします
        end_th_of_edge = 0
        for i in range(1, end_th_of_node):
            if i not in edge_th_set:
                break

            edge_th_set.remove(i)
            end_th_of_edge = i + 1

        # 残っている edge 列は others リストに入れます
        for edge_th in edge_th_set:
            others_name_list.append(f'edge{edge_th}')


        return is_n_o_column_exists, end_th_of_node, end_th_of_edge, others_name_list



import gc
import pandas as pd
from ...library import INDENT


##############
# MARK: Forest
##############
class Forest():
    """森"""


    def __init__(self):
        self._multiple_root_entry = {}

        self._order_of_remainder_columns = []

        # 探索時に使用する一時変数
        self._temp_leaf_th = None


    @property
    def multiple_root_entry(self):
        return self._multiple_root_entry


    @property
    def order_of_remainder_columns(self):
        """ツリー構造には含まないテーブル列の並び順を指定するのに使います"""
        return self._order_of_remainder_columns
    

    @order_of_remainder_columns.setter
    def order_of_remainder_columns(self, value):
        self._order_of_remainder_columns = value


    def tree_root(self, edge_text, node_text):
        """TODO 根ノードでのエッジテキストは未対応するか？"""
        root_entry = TreeEntry(parent_entry=None, edge_text=edge_text, node_text=node_text, child_entries={}, leaf_th=None)

        if root_entry._pack_key() in self._multiple_root_entry:
            raise ValueError(f"key exists  {root_entry._pack_key()=}")

        self._multiple_root_entry[root_entry._pack_key()] = root_entry

        return root_entry


    def renumbering(self):
        """番号の振り直し"""

        self._temp_leaf_th = 1

        for root_entry in self._multiple_root_entry.values():
            self.renumbering_child(root_entry)


    def renumbering_child(self, node):
        # 葉
        if not node.has_children():
            node.leaf_th = self._temp_leaf_th
            self._temp_leaf_th += 1
            return

        for child_entry in node._child_entries.values():
            self.renumbering_child(child_entry)  # 再帰


    def _stringify_like_tree(self, indent):
        items = []
        
        for root_entry in self._multiple_root_entry.values():
            items.append(root_entry._stringify_like_tree(indent=indent, as_root=True))

        return f"""\
{''.join(items)}"""


    def to_csv(self, csv_file_path):
        """CSV形式でファイルへ出力します"""

        # # 葉要素に番号を振っていく
        # self.renumbering()


        class Context():
            def __init__(self):
                self._cur_depth = 0
                self._max_depth = 0
                self._leaf_entries = []
                

        context = Context()


        def find_leaf(context, entry):
            """葉を収集する。葉の最大深さも調べる"""

            context._cur_depth += 1


            # 最大深さ
            if context._max_depth < context._cur_depth:
                context._max_depth = context._cur_depth


            # 葉要素
            if not entry.has_children():
                context._leaf_entries.append(entry)
            else:
                for child_entry in entry.child_entries.values():
                    find_leaf(context, child_entry) # 再帰


            context._cur_depth -= 1


        # 全ての葉を収集
        for root_entry in self.multiple_root_entry.values():
            find_leaf(context, root_entry)


        # 余り列の名前
        # ------------
        remainder_column_name_set = set()

        for leaf in context._leaf_entries:
            if leaf.remainder_columns is not None:
                for name, value in leaf.remainder_columns.items():
                    remainder_column_name_set.add(name)

        # 順序が指定されているものは消す
        for name in self.order_of_remainder_columns:
            if name in remainder_column_name_set:
                remainder_column_name_set.remove(name)

        # 順序を固定する
        order_of_remainder_columns = self.order_of_remainder_columns.copy()
        order_of_remainder_columns.extend(list(remainder_column_name_set))
        #print(f"余り列の順序指定：{self.order_of_remainder_columns=}")
        #print(f"余り列の名前　　：{order_of_remainder_columns=}")



        #print(f"最大深さ：{context._max_depth=}")

        # 出力する順番に列名を並べる（存在しない列が含まれても構わない。存在する列が含まれていなくても構わない）
        order_of_column_names = ['no']

        for i in range(0, context._max_depth + 1):
            order_of_column_names.append(f'edge{i}')
            order_of_column_names.append(f'node{i}')

        # 余り列を追加
        for remainder_column_name in order_of_remainder_columns:
            order_of_column_names.append(remainder_column_name)

        # print(f"列名の並び順：{order_of_column_names=}")
        # print(f"列名の並び順の要素数：{len(order_of_column_names)=}")

        df = pd.DataFrame()


        # 葉のすべての親を出力
        for leaf_th, leaf in enumerate(context._leaf_entries, 1):


            cur_entry = leaf
            path = [cur_entry]

            while cur_entry.parent_entry is not None:
                # 逆順で親エントリーが入っていく
                cur_entry = cur_entry.parent_entry
                path.append(cur_entry)


            record = {'no':leaf_th}

            # * `entry_no` - 根を 0 とする連番
            for entry_no, entry in enumerate(reversed(path)):
                if entry.edge_text is not None:
                    record[f'edge{entry_no}'] = entry.edge_text
                
                if entry.node_text is not None:
                    record[f'node{entry_no}'] = entry.node_text

            # 余り列を追加
            if leaf.remainder_columns is not None:
                for name, value in leaf.remainder_columns.items():
                    record[name] = value

#                   print(f"""\
# df:
# {df}
# {leaf_th=}
# 列名の並び順の要素数：{len(order_of_column_names[1:])=}
# 列名の並び順　　　　：{order_of_column_names[1:]=}
# レコード　　　　　　：{record}
# """)

            # テーブルに存在しない列は追加する
            for column_name in record.keys():
                if column_name not in df.columns.values:
                    df[column_name] = None


            def insert_record(df, leaf_th, record):
                """データフレームにレコード追加"""
                df.loc[leaf_th] = record


            #
            #   NOTE ここで、空テーブルや、空列と連結すると、 pandas から警告が出ることがある
            #   FutureWarning: The behavior of DataFrame concatenation with empty or all-NA entries is deprecated. In a future version, this will no longer exclude empty or all-NA columns when determining the result dtypes. To retain the old behavior, exclude the relevant entries before the concat operation.
            #
            # データフレームが空のとき
            if df.empty:
                insert_record(df=df, leaf_th=leaf_th, record=record)

            # データフレームが空でないとき
            else:
                insert_record(df=df, leaf_th=leaf_th, record=record)


        # 全部欠損している列を削除
        df.dropna(how='all', axis=1, inplace=True)


#         print(f"""\
# df:
# {df}
# 実際の列名の要素数　：{len(df.columns.values)}
# 実際の列名一覧　　　：{df.columns.values}
# 列名の並び順の要素数：{len(order_of_column_names[1:])}
# 列名の並び順　　　　：{order_of_column_names[1:]}
# """)

        # ［実際の列名一覧］をコピーして［順序の指定されていない列名一連］を作る
        no_order_of_column_names = list(df.columns.values)
        # print(f"(1) ［実際の列名一覧］={df.columns.values}")
        # print(f"(2) ［列名の並び順］={order_of_column_names}")

        # ［列名の並び順］に有る列名が、［順序の指定されていない列名一連］にあれば、［順序の指定されていない列名一連］から削除する
        for column_name in order_of_column_names:
            if column_name in no_order_of_column_names:
                no_order_of_column_names.remove(column_name)
        #print(f"(3) ［順序の指定されていない列名一連］={no_order_of_column_names}")

        # ［列名の並び順］から、［実際の列名一覧］に有る列名だけを残し、［再：列名の並び順］とする。このとき 'no' インデックスが消える
        reorder_of_column_names = []
        for column_name in order_of_column_names:
            if column_name in df.columns.values:
                reorder_of_column_names.append(column_name)
        #print(f"(4) ［再：列名の並び順］={reorder_of_column_names}")

        # ［再：列名の並び順］と［順序の指定されていない列名一連］を連結して、［出力する列名一連］とする
        output_column_names = reorder_of_column_names + no_order_of_column_names
        #print(f"(5) ［出力する列名一連］={output_column_names}")


#         print(f"""\
# f:
# {df}""")

        # テーブルをCSV形式でファイルへ保存
        df.to_csv(
                csv_file_path,
                encoding='utf8',
                columns=output_column_names,
                index=False)


        # オブジェクトの破棄
        del df
        # メモリ解放
        gc.collect()


#############
# MARK: Entry
#############
class TreeEntry():
    """ツリー・エントリー

    エッジとノードのペア
    
    イミュータブルにすると生成が難しいので、ミュータブルとする
    """


    def __init__(self, parent_entry, edge_text, node_text, child_entries, leaf_th=None, remainder_columns=None):
        """初期化
        
        Parameters
        ----------
        parent_entry : TreeEntry
            親ノード
        edge_text : str
            エッジのテキスト
        node_text : str
            ノードのテキスト
        child_entries : dict<tuple(str, str), TreeEntry>
            子ノードを格納した辞書。キーはエッジテキストとノードテキストのタプル
            FIXME キーがメモリを消費しすぎていないか？仕方ない？
        leaf_th : int
            有れば１から始まる葉番号、無ければナン
        remainder_columns : dict
            有れば、ツリー構造に含まれなかった列の辞書。無ければナン
        """
        self._parent_entry = parent_entry
        self._edge_text = edge_text
        self._node_text = node_text
        self._child_entries = child_entries
        self._leaf_th = leaf_th
        self._remainder_columns = remainder_columns


    @property
    def parent_entry(self):
        """親ノード"""
        return self._parent_entry


    @property
    def edge_text(self):
        """エッジのテキスト"""
        return self._edge_text


    @property
    def node_text(self):
        """ノードのテキスト"""
        return self._node_text


    @property
    def child_entries(self):
        """子ノードを格納した辞書。キーはエッジテキストとノードテキストのタプル
        FIXME キーがメモリを消費しすぎていないか？仕方ない？"""
        return self._child_entries


    @property
    def leaf_th(self):
        """有れば１から始まる葉番号、無ければナン"""
        return self._leaf_th


    @leaf_th.setter
    def leaf_th(self, value):
        """有れば１から始まる葉番号、無ければナン"""
        self._leaf_th = value


    @property
    def remainder_columns(self):
        """有れば、ツリー構造に含まれなかった列の辞書。無ければナン"""
        return self._remainder_columns


    @remainder_columns.setter
    def remainder_columns(self, value):
        """有れば、ツリー構造に含まれなかった列の辞書。無ければナン"""
        self._remainder_columns = value


    def has_children(self):
        """子エントリーを持つか？"""
        return 0 < len(self._child_entries)


    def leaf(self, edge_text, node_text, remainder_columns=None):
        """葉要素を生やします"""
        leaf_entry = self.grow(edge_text=edge_text, node_text=node_text)

        leaf_entry.remainder_columns = remainder_columns

        return leaf_entry


    def grow(self, edge_text, node_text):
        """子要素を生やします"""
        child_entry = TreeEntry(parent_entry=self, edge_text=edge_text, node_text=node_text, child_entries={})

        if child_entry._pack_key() in self._child_entries:
            raise ValueError(f"key exists  {child_entry._pack_key()=}")

        self._child_entries[child_entry._pack_key()] = child_entry

        return child_entry


    def get_child(self, edge_text, node_text, default=None):
        """子要素を取得。無ければデフォルト値を返します"""
        pack_key = TreeEntry._pack_key_static(edge_text=edge_text, node_text=node_text)
        if self.has_child(edge_text=edge_text, node_text=node_text):
            return self.child_entries[pack_key]
        else:
            return default


    def has_child(self, edge_text, node_text):
        """子要素が既存か？"""
        pack_key = TreeEntry._pack_key_static(edge_text=edge_text, node_text=node_text)
        return pack_key in self._child_entries


    @staticmethod
    def _pack_key_static(edge_text, node_text):
        return (edge_text, node_text)


    def _pack_key(self):
        return TreeEntry._pack_key_static(edge_text=self._edge_text, node_text=self._node_text)


    def _stringify_like_tree(self, indent, as_root=False):
        succ_indent = indent + INDENT


        if as_root:
            et = '──'
        else:
            et = '└─'


        if self._edge_text is not None:
            et += f"{self._edge_text}─"


        if not self.has_children():
            icon = f'📄 ({self._leaf_th})'
        else:
            icon = '📁'


        if self._remainder_columns is not None:
            remander_columns_text = f"  {self._remainder_columns}"
        else:
            remander_columns_text = ''


        items = []
        for child_entry in self._child_entries.values():
            items.append(child_entry._stringify_like_tree(indent=succ_indent))


        return f"""\
{indent}{et} {icon} {self._node_text}{remander_columns_text}
{''.join(items)}"""


    def _stringify_dump(self, indent):
        succ_indent = indent + INDENT

        items = []
        for child_entry in self._child_entries.values():
            items.append(child_entry._stringify_dump(indent=succ_indent))

        return f"""\
{indent}TreeEntry
{indent}---------
{succ_indent}{self._edge_text=}
{succ_indent}{self._node_text=}
{succ_indent}{self._remainder_columns=}
{''.join(items)}"""



from collections import deque
from . import Forest, TreeEntry


#######################
# REMARK: TreeStructure
#######################
class TreeStructureBasedOnTable():
    """テーブル構造を読み込んだ、マルチ根を持つ木構造"""


    @staticmethod
    def read_table_and_planting(table):
        """テーブル読取

        Parameters
        ----------
        table : Table
            データテーブル

        Returns
        -------
        forest : Forest
            森
        """

        tree_structure = TreeStructureBasedOnTable()

        # 先頭のレコードから順に読み込んでいけば作れる
        table.for_each(tree_structure.on_record_read)

        # # ダンプ
        # print("[read] マルチ根")
        # for root in tree_structure._forest.multiple_root_entry.values():
        #     print(f"{root._stringify_like_tree('    ')}")

        return tree_structure._forest


    def __init__(self):
        """初期化"""

        # 森
        self._forest = Forest()


    def on_record_read(self, row_number, record):
        """レコード読取
        
        例えば ^ をマルチ根の親とするとき、
        ^, A, B, C
        ^, A, B, D
        という２レコードを読込むとき、 
        ２行目では A, B というパスは既存。

        """

        class Context():
            def __init__(self):
                self._stack = deque()
                self._pre_parent_tree_entry = None
        
        context = Context()


        def set_node(self, context, leaf_th, depth, node_in_record):
#             print(f"""[set_node] {depth=}
# {node_in_record._stringify_dump('')}""")

            # 既存のマルチ根かもしれない
            if depth==0 and node_in_record.text in self._forest.multiple_root_entry:
                tree_entry = self._forest.multiple_root_entry[node_in_record.text]

            # 未作成のノードなら
            elif context._pre_parent_tree_entry is None or node_in_record._pack_key() not in context._pre_parent_tree_entry.child_entries:
                tree_entry = TreeEntry(
                        parent_entry=context._pre_parent_tree_entry,
                        edge_text=node_in_record.edge_text,
                        node_text=node_in_record.text,
                        child_entries={},     # 子要素は、子から戻ってきたときじゃないと分からない
                        leaf_th=leaf_th)
            
            # 既存のノードなら
            else:
                tree_entry = context._pre_parent_tree_entry.child_entries[node_in_record._pack_key()]


            context._pre_parent_tree_entry = tree_entry
            context._stack.append(tree_entry)


        # if row_number == 0:
        #     print("最初のレコードは、根ノードから葉ノードまで全部揃ってる")

        def get_leaf_th(record, depth):
            if depth<record.len_of_path_from_root_to_leaf:
                return record.no
            return None

        record.for_each_node_in_path(set_node=lambda depth, node_in_record: set_node(self=self, context=context, leaf_th=get_leaf_th(record=record, depth=depth), depth=depth, node_in_record=node_in_record))


        prev_child_tree_entry = None

        # 葉から根に向かってノードを読取
        while 0 < len(context._stack):
            tree_entry = context._stack.pop()

            # 子を、子要素として追加
            if prev_child_tree_entry is not None:
                tree_entry.child_entries[prev_child_tree_entry._pack_key()] = prev_child_tree_entry

            #print(f"逆読み  {tree_entry.edge_text=}  {tree_entry.node_text=}")
            prev_child_tree_entry = tree_entry


        if len(context._stack) != 0:
            raise ValueError(f"スタックのサイズが0でないのはおかしい  {len(context._stack)=}")


        # ルートノードを記憶
        self._forest.multiple_root_entry[tree_entry.node_text] = tree_entry


#         print(f"""レコード読取  {row_number=}
# root_entry:
# {tree_entry._stringify_dump('')}
# record:
# {record._stringify_dump('')}""")



import gc
import datetime
from ..models.database import Table
from ..models.tree_structure.converter import TreeStructureBasedOnTable
from ..settings import Settings
from .draw_tools import TreeDrawer, TreeEraser


class WorksheetHandle():
    """ワークシートへの対応"""


    @staticmethod
    def instantiate(target, based_on, ws, settings_obj, debug_write=False):
        """生成

        Parameters
        ----------
        target : str
            シート名
        based_on : str
            CSVファイルパス
        ws : Worksheet
            ワークシート
        settings_obj : Settings
            各種設定
        debug_write : bool
            デバッグライト
        """

        # CSV読込
        table = Table.from_csv(file_path=based_on)

        # table からツリー構造を作成
        #
        #   NOTE マルチ根にも対応していることに注意してください
        #
        forest = TreeStructureBasedOnTable.read_table_and_planting(table=table)

        return WorksheetHandle(target=target, based_on=based_on, ws=ws, settings_obj=settings_obj, table=table, forest=forest, debug_write=debug_write)


    def __init__(self, target, based_on, ws, settings_obj, table, forest, debug_write=False):
        """初期化

        Parameters
        ----------
        target : str
            シート名
        based_on : str
            CSVファイルパス
        ws : Worksheet
            ワークシート
        settings_obj : Settings
            各種設定
        table : .models.database.Table
            データテーブル
        forest : Forest
            森
        debug_write : bool
            デバッグライト
        """
        self._target = target
        self._based_on = based_on
        self._ws = ws
        self._settings_obj = settings_obj
        self._table = table
        self._forest = forest
        self._debug_write = debug_write


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        del self._target
        del self._based_on
        del self._ws
        del self._settings_obj
        del self._table
        del self._forest
        del self._debug_write

        # メモリ解放
        gc.collect()


    @property
    def forest(self):
        """森"""
        return self._forest


    def render_tree(self):
        """木の描画"""

        # ツリードロワーを用意、描画（都合上、要らない罫線が付いています）
        tree_drawer = TreeDrawer(table=self._table, ws=self._ws, settings_obj=self._settings_obj, debug_write=self._debug_write)
        tree_drawer.render()


        # 要らない罫線を消す
        # DEBUG_TIPS: このコードを不活性にして、必要な線は全部描かれていることを確認してください
        if True:
            tree_eraser = TreeEraser(table=self._table, ws=self._ws, settings_obj=self._settings_obj, debug_write=self._debug_write)
            tree_eraser.render()
        else:
            if self._debug_write:
                print(f"[{datetime.datetime.now()}] eraser disabled")



import datetime
import pandas as pd
from .library import TableControl


class ColumnsSorting():
    """列ソート、及びその途中の計算結果"""


    def __init__(self):
        # 計算途中結果
        self._is_n_o_column_exists = None
        self._end_th_of_node = None
        self._end_th_of_edge = None


    @property
    def end_th_of_node(self):
        """node0 から連続する最後の node{i} の i を +1 したもの。なければ 0"""
        return self._end_th_of_node


    @property
    def end_th_of_edge(self):
        """edge1 から ［edge{end_th_of_node} の手前］まで連続する最後の edge{i} の i を +1 したもの。なければ 0"""
        return self._end_th_of_edge


    def execute(self, df):
        """実行

        列名をソートする。no,node0,edge1,node1,edge2,node2,remaining_a,remaining_b,... のような感じに

        'no' 列は、（まだインデックスに指定されていないものとし）有るケースと、無いケースがあります。有れば先頭へ、無ければ無視することにします
        """

        self._is_n_o_column_exists, self._end_th_of_node, self._end_th_of_edge, others_name_list = TableControl.sort_out_column_names_n_o_node_edge_others(df)

        column_name_list = []

        # 'no' 列を含むなら先頭へ
        if self._is_n_o_column_exists:
            column_name_list.append('no')

        if self._end_th_of_node < 1:
            raise ValueError(f'node0 列がありませんでした  {self._end_th_of_edge=}  {self._end_th_of_node=}  {others_name_list=}')
        
        # node0 列を追加
        column_name_list.append('node0')

        for i in range(1, self._end_th_of_node):

            # あれば edge{i} 列を追加
            if i < self._end_th_of_edge:
                column_name_list.append(f'edge{i}')

            # node{i} 列を追加
            column_name_list.append(f'node{i}')

        # 残りの列名を追加
        column_name_list.extend(others_name_list)

        return df[column_name_list]


class InputCompletion():
    """入力補完"""


    @staticmethod
    def fill_directory(df, end_th_of_edge, end_th_of_node, debug_write=False):
        """ディレクトリーの空欄を埋めます
        
        対象は、エッジテキスト、ノードテキスト列です

        Before:
            a,b,c,d,e,f,g,h,i
                ,j,k,l, ,m,n,o
                , ,p,      q
        
        After:
            a,b,c,d,e,f,g,h,i
            a,j,k,l,e,m,n,o
            a,j,p,l,e,m,n
        """

        if debug_write:
            print(f"[{datetime.datetime.now()}] このテーブルは{end_th_of_node}個のノードがある。最終ノードは {end_th_of_node - 1}")

        row_size = len(df)

        # ２行目から、１行ずつ行う
        for row_th in range(2, row_size + 1):

            InputCompletion.copy(df=df, row_th=row_th, start_column_th=1, prefix='edge', end_th=end_th_of_edge, debug_write=debug_write)
            InputCompletion.copy(df=df, row_th=row_th, start_column_th=0, prefix='node', end_th=end_th_of_node, debug_write=debug_write)


    @staticmethod
    def copy(df, row_th, start_column_th, prefix, end_th, debug_write):

        # この行について、最終ノード列を調べる
        actual_last_th = end_th - 1   # 最終ノードから開始
        for element_th in reversed(range(start_column_th, end_th)):
            column_name = f'{prefix}{element_th}'

            # 縮めていく
            actual_last_th = element_th

            if not pd.isnull(df.at[row_th, column_name]):
                break


        if debug_write:
            print(f"[{datetime.datetime.now()}] 第{row_th}行の{prefix}は第{actual_last_th}要素まで")

        # この行について、最終要素列まで、要素の空欄は上行をコピーする
        for element_th in range(start_column_th, actual_last_th + 1):

            column_name = f'{prefix}{element_th}'

            if pd.isnull(df.at[row_th, column_name]):
                df.at[row_th, column_name] = df.at[row_th - 1, column_name]



from .library import TableControl


class SourceCsvTableAnalyzer():


    def __init__(self, df, end_th_of_edge, end_th_of_node):
        self._df = df
        self._end_th_of_edge = end_th_of_edge
        self._end_th_of_node = end_th_of_node


    @staticmethod
    def instantiate(df, end_th_of_edge, end_th_of_node):
        return SourceCsvTableAnalyzer(df=df, end_th_of_edge=end_th_of_edge, end_th_of_node=end_th_of_node)


    @property
    def end_th_of_edge(self):
        return self._end_th_of_edge


    @property
    def end_th_of_node(self):
        return self._end_th_of_node


    def get_column_name_of_last_node(self):
        """最終ノードの列名"""
        return f'node{self._end_th_of_node - 1}'


    def get_column_th_of_last_node(self):
        """最終ノードの列番号"""
        return TableControl.get_column_location_by_column_name(df=self._df, column_name=self.get_column_name_of_last_node()) + 1



from openpyxl.styles import PatternFill, Font
from openpyxl.styles.borders import Border, Side
from openpyxl.styles.alignment import Alignment


class Settings():
    """各種設定"""


    def __init__(self, dictionary=None):
        """初期化
        
        Parameters
        ----------
        dictionary : dict
            設定

            列の幅設定。width はだいたい 'ＭＳ Ｐゴシック' サイズ11 の半角英文字の個数
            * `column_width_of_no` - A列の幅。no列
            * `column_width_of_root_side_padding` - B列の幅。ツリー構造図の根側パディング
            * `column_width_of_leaf_side_padding` - ツリー構造図の葉側パディング
            * `column_width_of_node` - 例：C, F, I ...列の幅。ノードの箱の幅
            * `column_width_of_parent_side_edge` - 例：D, G, J ...列の幅。エッジの水平線のうち、親ノードの方
            * `column_width_of_child_side_edge` - 例：E, H, K ...列の幅。エッジの水平線のうち、子ノードの方

            行の高さ設定。height の単位はポイント。既定値 8。昔のアメリカ人が椅子に座ってディスプレイを見たとき 1/72 インチに見える大きさが 1ポイント らしいが、そんなんワカラン。目視確認してほしい
            * `row_height_of_header` - 第１行。ヘッダー
            * `row_height_of_lower_side_padding` - 第２行。ツリー構造図の軸の番号が小さい側パティング
            * `row_height_of_upper_side_of_node` - ノードの上側のセルの高さ
            * `row_height_of_lower_side_of_node` - ノードの下側のセルの高さ
            * `row_height_of_node_spacing` - ノード間の高さ

            * 色の参考： 📖 [Excels 56 ColorIndex Colors](https://www.excelsupersite.com/what-are-the-56-colorindex-colors-in-excel/)

            背景色関連
            * `bgcolor_of_tree` - ツリー構造図の背景
            * `bgcolor_of_header_1` - ヘッダーの背景色その１
            * `bgcolor_of_header_2` - ヘッダーの背景色その２
            * `bgcolor_of_node` - 背景色

            文字色関連
            * `fgcolor_of_header_1` - ヘッダーの文字色その１
            * `fgcolor_of_header_2` - ヘッダーの文字色その２

            文字寄せ関連
            * `horizontal_alignment_of_node` - 文字の水平方向の寄せ。規定値 None。'left', 'fill', 'centerContinuous', 'center', 'right', 'general', 'justify', 'distributed' のいずれか。指定しないなら None
            * `vertical_alignment_of_node` - 文字の垂直方向の寄せ。規定値 None。'bottom', 'center', 'top', 'justify', 'distributed' のいずれか。指定しないなら None

            その他の操作
            * `do_not_merge_cells` - セル結合しないなら真
        """

        # 既定のディクショナリー
        # いわゆる settings
        self._dictionary = {
            # 列の幅
            #
            #   ［列幅の自動調整］機能を付けたので、文字が入る箇所は規定値はナンにします。
            #   キーは存在させたいので、コメントアウトしないでください
            #
            'column_width_of_no':                    None,
            'column_width_of_root_side_padding':        3,
            'column_width_of_leaf_side_padding':        3,
            'column_width_of_node':                  None,
            'column_width_of_parent_side_edge':         2,
            'column_width_of_child_side_edge':       None,

            # 行の高さ
            'row_height_of_header':                    13,
            'row_height_of_lower_side_padding':        13,
            'row_height_of_upper_side_of_node':        13,
            'row_height_of_lower_side_of_node':        13,
            'row_height_of_node_spacing':               6,

            # 背景色関連
            'bgcolor_of_tree':                   'FFFFFF',
            'bgcolor_of_header_1':               'CCCCCC',
            'bgcolor_of_header_2':               '333333',
            'bgcolor_of_node':                   'FFFFCC',

            # 文字色関連
            'fgcolor_of_header_1':               '111111',
            'fgcolor_of_header_2':               'EEEEEE',

            # 文字寄せ関連
            'horizontal_alignment_of_node':          None,
            'vertical_alignment_of_node':            None,

            # その他の操作
            'do_not_merge_cells':                   False,      # セル結合しないなら真
        }

        # 上書き
        if dictionary is not None:
            for key, value in dictionary.items():
                self._dictionary[key] = value


        # フォント関連
        # ------------
        self._font_of_header_list = []

        color = self.dictionary['fgcolor_of_header_1']
        if color is not None:
            self._font_of_header_list.append(Font(color=color))
        
        color = self.dictionary['fgcolor_of_header_2']
        if color is not None:
            self._font_of_header_list.append(Font(color=color))


        # 罫線関連
        # --------

        # エッジの罫線
        #
        #   style に入るもの： 'dashDot', 'dashDotDot', 'double', 'hair', 'dotted', 'mediumDashDotDot', 'dashed', 'mediumDashed', 'slantDashDot', 'thick', 'thin', 'medium', 'mediumDashDot'
        #   色の参考： 📖 [Excels 56 ColorIndex Colors](https://www.excelsupersite.com/what-are-the-56-colorindex-colors-in-excel/)
        #
        BLACK = '000000'
        side = Side(style='thick', color=BLACK)

        # DEBUG_TIPS: 罫線に色を付けると、デバッグしやすいです
        if True:
            red_side = Side(style='thick', color=BLACK)
            orange_side = Side(style='thick', color=BLACK)
            green_side = Side(style='thick', color=BLACK)
            blue_side = Side(style='thick', color=BLACK)
            cyan_side = Side(style='thick', color=BLACK)
        else:
            red_side = Side(style='thick', color='FF0000')
            orange_side = Side(style='thick', color='FFCC00')
            green_side = Side(style='thick', color='00FF00')
            blue_side = Side(style='thick', color='0000FF')
            cyan_side = Side(style='thick', color='00FFFF')

        # ─字  赤
        self._border_to_parent_horizontal = Border(bottom=red_side)
        self._under_border_to_child_horizontal = Border(bottom=red_side)
        # │字  緑
        self._leftside_border_to_vertical = Border(left=green_side)
        # ┬字  青
        self._border_to_parent_downward = Border(bottom=blue_side)
        self._under_border_to_child_downward = Border(bottom=blue_side)
        self._leftside_border_to_child_downward = Border(left=blue_side)
        # ├字  青緑
        self._l_letter_border_to_child_rightward = Border(left=cyan_side, bottom=cyan_side)
        self._leftside_border_to_child_rightward = Border(left=cyan_side)
        # └字  橙
        self._l_letter_border_to_child_upward = Border(left=orange_side, bottom=orange_side)

        # DEBUG_TIPS: デバッグ時は、罫線を消すのではなく、灰色に変えると見やすいです
        if True:
            # 罫線無し
            self._striked_border = None
        else:
            # 罫線
            #
            #   style に入るもの： 'dashDot', 'dashDotDot', 'double', 'hair', 'dotted', 'mediumDashDotDot', 'dashed', 'mediumDashed', 'slantDashDot', 'thick', 'thin', 'medium', 'mediumDashDot'
            #   色の参考： 📖 [Excels 56 ColorIndex Colors](https://www.excelsupersite.com/what-are-the-56-colorindex-colors-in-excel/)
            #
            # 見え消し用（デバッグに使う）
            striked_side = Side(style='thick', color='DDDDDD')
            # 見え消し用の罫線
            self._striked_border = Border(left=striked_side)


        # ノードの罫線
        #
        #   style に入るもの： 'dashDot', 'dashDotDot', 'double', 'hair', 'dotted', 'mediumDashDotDot', 'dashed', 'mediumDashed', 'slantDashDot', 'thick', 'thin', 'medium', 'mediumDashDot'
        #
        side = Side(style='thick', color='000000')
        self._border_of_upside_node = Border(top=side, left=side, right=side)
        self._border_of_downside_node = Border(bottom=side, left=side, right=side)


        # 余り列
        side = Side(style='thin', color='111111')
        self._upper_border_of_remaining_cell = Border(top=side, left=side, right=side)
        self._middle_border_of_remaining_cell = Border(left=side, right=side)
        self._lower_border_of_remaining_cell = Border(bottom=side, left=side, right=side)


        # 背景色関連
        # ----------
        self._list_of_bgcolor_of_header = []

        color = self.dictionary['bgcolor_of_header_1']
        if color is not None:
            self._list_of_bgcolor_of_header.append(PatternFill(patternType='solid', fgColor=color))
        else:
            self._list_of_bgcolor_of_header.append(None)
        
        color = self.dictionary['bgcolor_of_header_2']
        if color is not None:
            self._list_of_bgcolor_of_header.append(PatternFill(patternType='solid', fgColor=color))
        else:
            self._list_of_bgcolor_of_header.append(None)

        # ツリー構造図の背景色
        color = self.dictionary['bgcolor_of_tree']
        if color is not None:
            self._bgcolor_of_tree = PatternFill(patternType='solid', fgColor=color)

        # ノードの背景色
        color = self.dictionary['bgcolor_of_node']
        if color is not None:
            self._bgcolor_of_node = PatternFill(patternType='solid', fgColor=color)


        # 文字寄せ関連
        # ------------
        #
        #   horizontal は 'distributed', 'fill', 'general', 'center', 'centerContinuous', 'justify', 'right', 'left' のいずれかから選ぶ
        #   vertical は 'center', 'top', 'bottom', 'justify', 'distributed' のいずれかから選ぶ
        #   TODO Excelでは文字は left、数は right など規定値が型によって違うが、そこまで細かく設定を分けてない。分けるか？
        #
        self._alignment_left_center = Alignment(horizontal='left', vertical='center')
        self._alignment_center_center = Alignment(horizontal='center', vertical='center')

        horizontal = self.dictionary['horizontal_alignment_of_node']
        vertical = self.dictionary['vertical_alignment_of_node']
        if horizontal is not None and vertical is not None:
            self._alignment_of_node = Alignment(horizontal=horizontal, vertical=vertical)
        elif horizontal is not None:
            self._alignment_of_node = Alignment(horizontal=horizontal)
        elif vertical is not None:
            self._alignment_of_node = Alignment(vertical=vertical)
        else:
            self._alignment_of_node = None


    @property
    def dictionary(self):
        return self._dictionary


    def set_alignment_left_center(self, cell):
        cell.alignment = self._alignment_left_center


    def set_alignment_center_center(self, cell):
        cell.alignment = self._alignment_center_center


    def set_striked_border(self, cell):
        # FIXME None にするという動作。どう対称性を取る？
        #if self._striked_border is not None:
        cell.border = self._striked_border


    def set_border_of_upside_node(self, cell):
        if self._border_of_upside_node is not None:
            cell.border = self._border_of_upside_node


    def set_border_of_downside_node(self, cell):
        if self._border_of_downside_node is not None:
            cell.border = self._border_of_downside_node


    def set_border_to_parent_horizontal(self, cell):
        if self._border_to_parent_horizontal is not None:
            cell.border = self._border_to_parent_horizontal


    def set_under_border_to_child_horizontal(self, cell):
        if self._under_border_to_child_horizontal is not None:
            cell.border = self._under_border_to_child_horizontal


    def set_leftside_border_to_vertical(self, cell):
        if self._leftside_border_to_vertical is not None:
            cell.border = self._leftside_border_to_vertical


    def set_border_to_parent_downward(self, cell):
        if self._border_to_parent_downward is not None:
            cell.border = self._border_to_parent_downward


    def set_under_border_to_child_downward(self, cell):
        if self._under_border_to_child_downward is not None:
            cell.border = self._under_border_to_child_downward


    def set_leftside_border_to_child_downward(self, cell):
        if self._leftside_border_to_child_downward is not None:
            cell.border = self._leftside_border_to_child_downward


    def set_l_letter_border_to_child_rightward(self, cell):
        if self._l_letter_border_to_child_rightward is not None:
            cell.border = self._l_letter_border_to_child_rightward


    def set_leftside_border_to_child_rightward(self, cell):
        if self._leftside_border_to_child_rightward is not None:
            cell.border = self._leftside_border_to_child_rightward


    def set_l_letter_border_to_child_upward(self, cell):
        if self._l_letter_border_to_child_upward is not None:
            cell.border = self._l_letter_border_to_child_upward


    def set_upper_border_of_remaining_cell(self, cell):
        if self._upper_border_of_remaining_cell is not None:
            cell.border = self._upper_border_of_remaining_cell


    def set_middle_border_of_remaining_cell(self, cell):
        if self._middle_border_of_remaining_cell is not None:
            cell.border = self._middle_border_of_remaining_cell


    def set_lower_border_of_remaining_cell(self, cell):
        if self._lower_border_of_remaining_cell is not None:
            cell.border = self._lower_border_of_remaining_cell


    def set_bgcolor_of_header_to(self, cell, index):
        if self._list_of_bgcolor_of_header[index] is not None:
            cell.fill = self._list_of_bgcolor_of_header[index]


    def set_font_of_header_to(self, cell, index):
        if self._font_of_header_list[index] is not None:
            cell.font = self._font_of_header_list[index]


    def set_bgcolor_of_tree_to(self, cell):
        if self._bgcolor_of_tree is not None:
            cell.fill = self._bgcolor_of_tree


    def set_bgcolor_of_node_to(self, cell):
        if self._bgcolor_of_node is not None:
            cell.fill = self._bgcolor_of_node


    def set_alignment_of_node_to(self, cell):
        if self._alignment_of_node is not None:
            cell.alignment = self._alignment_of_node


class SettingsOfNode():
    """TODO ノード個別の設定"""


    def __init__(self, dictionary=None):
        """初期化
        
        Parameters
        ----------
        dictionary : dict
            設定

            * 色の参考： 📖 [Excels 56 ColorIndex Colors](https://www.excelsupersite.com/what-are-the-56-colorindex-colors-in-excel/)

            色関連
            * `bgcolor` - 背景色
            * `fgcolor` - 文字色

            文字寄せ関連
            * `horizontal_alignment` - 文字の水平方向の寄せ。規定値 None。'left', 'fill', 'centerContinuous', 'center', 'right', 'general', 'justify', 'distributed' のいずれか。指定しないなら None
            * `vertical_alignment_` - 文字の垂直方向の寄せ。規定値 None。'bottom', 'center', 'top', 'justify', 'distributed' のいずれか。指定しないなら None
        """

        # 既定のディクショナリー
        self._dictionary = {

            # 色関連
            'bgcolor':                     'FFFFCC',
            'fgcolor':                     None,

            # 文字寄せ関連
            'horizontal_alignment':        None,
            'vertical_alignment':          None,
        }

        # 上書き
        if dictionary is not None:
            for key, value in dictionary.items():
                self._dictionary[key] = value


    @property
    def dictionary(self):
        return self._dictionary



import datetime
import pandas as pd
import openpyxl as xl
from ..library import nth
from ..models.database import NodeInRecord, Record
from ..models.database.library import TableControl
from ..models.line_by_line_leaf import LineByLineLeafModel
from .style import StyleControl


class TreeDrawer():
    """エクセルで罫線などを駆使して、樹形図を描画します"""


    def __init__(self, table, ws, settings_obj, debug_write=False):
        """初期化
        
        Parameters
        ----------
        table : Table
            ツリーテーブル
        ws : openpyxl.Worksheet
            ワークシート
        settings_obj : Settings
            各種設定
        debug_write : bool
            デバッグライト
            DEBUG_TIPS: デバッグライトをオンにして、コンソールにログを表示すると不具合を調査しやすくなります
        """
        self._table = table
        self._ws = ws
        self._settings_obj = settings_obj
        self._debug_write = debug_write

        self._prev_record = Record.new_empty(specified_end_th_of_node=self._table.analyzer.end_th_of_node)
        self._curr_record = Record.new_empty(specified_end_th_of_node=self._table.analyzer.end_th_of_node)
        self._next_record = Record.new_empty(specified_end_th_of_node=self._table.analyzer.end_th_of_node)


    def render(self):
        """描画"""

        # 列幅の自動調整
        # --------------
        #
        #   NOTE 'no' は列ではなくインデックスなのでやり方が異なる
        #   NOTE 文字数は取れるが、１文字の横幅が１とは限らない
        #
        # 'no' インデックス
        width = len('no')
        for index_value in self._table.df.index:
            width = max(width, len(str(index_value)))
        self._ws.column_dimensions['A'].width = width * 1.5

        # エッジ列が省略されているケース
        #
        #   NOTE 規定値を何か設定しないと、デフォルトの長い列幅になってしまう
        #
        for source_column_th, column_name in enumerate(self._table.df.columns, 1):

            # ノードを見つければ、その前列がエッジ
            result = TableControl.pattern_of_column_name_of_node.match(column_name)
            if result:
                target_column_th = StyleControl.get_target_column_th(source_table=self._table, column_name=column_name)
                target_column_letter = xl.utils.get_column_letter(target_column_th - 1)
                # 幅は 4 でいいだろ
                self._ws.column_dimensions[target_column_letter].width = 4
                continue

        # 各列
        for source_column_th, column_name in enumerate(self._table.df.columns, 1):
            target_column_th = StyleControl.get_target_column_th(source_table=self._table, column_name=column_name)
            target_column_letter = xl.utils.get_column_letter(target_column_th)
            number_of_character = StyleControl.get_number_of_character_of_column(df=self._table.df, column_name=column_name)
            # Root とか 1st とか、ヘッダーに入っているので、とりあえず最低でも 4文字は欲しい。 1000th とかはとりあえず考えないことにする
            number_of_character = max(number_of_character, 4)

            #print(f"列幅の自動調整  {column_name=}  {target_column_letter=}  内訳：  {source_column_th=}  {target_column_th=}  {number_of_character=}")

            # ノード
            result = TableControl.pattern_of_column_name_of_node.match(column_name)
            if result:
                # 文字幅を 1.2 倍 + 1 ぐらいしておく
                # FIXME フォント情報からきっちり横幅を取れないか？
                self._ws.column_dimensions[target_column_letter].width = number_of_character * 1.2 + 1
                continue


            # エッジ
            #
            #   余白を開けたいから広くとる
            #
            result = TableControl.pattern_of_column_name_of_edge.match(column_name)
            if result:
                # 文字幅の 1.2 倍 + 4 ぐらいしておく
                # FIXME フォント情報からきっちり横幅を取れないか？
                self._ws.column_dimensions[target_column_letter].width = number_of_character * 1.2 + 4
                continue

            # 余り列
            # ------
            # 余り情報だし、余白は要らないから、文字幅を 1.2 倍ぐらいしておく
            # FIXME フォント情報からきっちり横幅を取れないか？
            self._ws.column_dimensions[target_column_letter].width = number_of_character * 1.2


        # 対象シートへ列ヘッダー書出し
        self._on_header()

        # 対象シートへの各行書出し
        self._table.for_each(on_each=self._on_each_record)

        # 最終行の実行
        self._on_each_record(next_row_number=len(self._table.df), next_record=Record.new_empty(specified_end_th_of_node=self._table.analyzer.end_th_of_node))

        # ウィンドウ枠の固定
        self._ws.freeze_panes = 'B2'


    def _forward_cursor(self, next_record):
        """送り出し

        Parameters
        ----------
        next_record : Record
            次行
        """
        self._prev_record = self._curr_record
        self._curr_record = self._next_record
        self._next_record = next_record


    def _on_header(self):

        # 変数名の短縮
        ws = self._ws


        # 列の幅設定
        column_width_dict = {}

        width = self._settings_obj.dictionary['column_width_of_no']
        if width is not None:
            column_width_dict['A'] = width      # no

        width = self._settings_obj.dictionary['column_width_of_root_side_padding']
        if width is not None:
            column_width_dict['B'] = width      # B列の幅。ツリー構造図の根側パディング
        
        width = self._settings_obj.dictionary['column_width_of_node']
        if width is not None:
            column_width_dict['C'] = width      # 根


        head_column_th = 4
        for node_th in range(1, self._table.analyzer.end_th_of_node):

            width = self._settings_obj.dictionary['column_width_of_parent_side_edge']
            if width is not None:
                column_width_dict[xl.utils.get_column_letter(head_column_th    )] = width   # 第n層  親側辺

            width = self._settings_obj.dictionary['column_width_of_child_side_edge']
            if width is not None:
                column_width_dict[xl.utils.get_column_letter(head_column_th + 1)] = width   #        子側辺

            width = self._settings_obj.dictionary['column_width_of_node']
            if width is not None:
                column_width_dict[xl.utils.get_column_letter(head_column_th + 2)] = width   #        節

            head_column_th += 3


        for name, width in column_width_dict.items():
            ws.column_dimensions[name].width = width


        # 行の高さ設定
        # height の単位はポイント。初期値 8。昔のアメリカ人が椅子に座ってディスプレイを見たとき 1/72 インチに見える大きさが 1ポイント らしいが、そんなんワカラン。目視確認してほしい
        row_height_dict = {}

        height = self._settings_obj.dictionary['row_height_of_header']
        if height is not None:
            row_height_dict[1] = height

        height = self._settings_obj.dictionary['row_height_of_lower_side_padding']
        if height is not None:
            row_height_dict[2] = height

        for row_number, height in row_height_dict.items():
            ws.row_dimensions[row_number].height = height


        # 第１行
        # ------
        # ヘッダー行にする
        row_th = 1

        # TODO 可変長ノード数への対応
        # NOTE データテーブルではなく、ビュー用途なので、テーブルとしての機能性は無視しています
        # A の代わりに {xl.utils.get_column_letter(1)} とも書ける
        ws[f'A{row_th}'] = 'No'
        self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'A{row_th}'], index=0)
        self._settings_obj.set_font_of_header_to(cell=ws[f'A{row_th}'], index=0)

        # 根側パディング
        # --------------
        self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'B{row_th}'], index=1)

        # 根
        # --
        ws[f'C{row_th}'] = 'Root'
        self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'C{row_th}'], index=1)
        self._settings_obj.set_font_of_header_to(cell=ws[f'C{row_th}'], index=1)


        flip = 0
        head_column_th = 4

        for node_th in range(1, self._table.analyzer.end_th_of_node):
            # 背景色、文字色
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'{xl.utils.get_column_letter(head_column_th    )}{row_th}'], index=flip)
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'{xl.utils.get_column_letter(head_column_th + 1)}{row_th}'], index=flip)
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'{xl.utils.get_column_letter(head_column_th + 2)}{row_th}'], index=flip)
            self._settings_obj.set_font_of_header_to(cell=ws[f'{xl.utils.get_column_letter(head_column_th + 2)}{row_th}'], index=flip)

            # 列名
            ws[f'{xl.utils.get_column_letter(head_column_th + 2)}{row_th}'] = nth(node_th)

            flip = (flip + 1) % 2
            head_column_th += 3


        # 葉側パディング
        # --------------
        target_column_th = self._table.analyzer.end_th_of_node * StyleControl.ONE_NODE_COLUMNS + 1
        column_letter = xl.utils.get_column_letter(target_column_th)
        cell_address = f'{column_letter}{row_th}'
        # 背景色、文字色
        self._settings_obj.set_bgcolor_of_header_to(cell=ws[cell_address], index=(flip + 1) % 2)   # 葉ノードと同じ色にする
        self._settings_obj.set_font_of_header_to(cell=ws[cell_address], index=(flip + 1) % 2)

        width = self._settings_obj.dictionary['column_width_of_leaf_side_padding']
        if width is not None:
            ws.column_dimensions[column_letter].width = width


        # 余り列
        # ------
        # 最終層以降の列
        column_name_of_leaf_entry = self._table.analyzer.get_column_name_of_last_node()
        is_remaining = False
        target_column_th = self._table.analyzer.end_th_of_node * StyleControl.ONE_NODE_COLUMNS + 2   # 空列を１つ挟む
        for column_name in self._table.df.columns:

            # ツリー区は無視
            if column_name == column_name_of_leaf_entry:
                #print(f'ツリー区 {row_th=}  {column_name=}')
                is_remaining = True
                continue

            elif is_remaining:
                cell_address = f'{xl.utils.get_column_letter(target_column_th)}{row_th}'
                #print(f'{cell_address=}  {row_th=}  {column_name=}')

                # 列名
                ws[cell_address].value = column_name
                # 背景色、文字色
                self._settings_obj.set_bgcolor_of_header_to(cell=ws[cell_address], index=flip)   # 葉ノードと同じ色にする
                self._settings_obj.set_font_of_header_to(cell=ws[cell_address], index=flip)

                flip = (flip + 1) % 2
                target_column_th += 1


        # 第２行
        # ------
        # 空行にする
        row_th = 2
        self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'A{row_th}'], index=0)   # 葉ノードと同じ色にする

        # ツリー構造図の背景色
        for column_th in range(2, target_column_th):
            column_letter = xl.utils.get_column_letter(column_th)
            self._settings_obj.set_bgcolor_of_tree_to(cell=ws[f'{column_letter}{row_th}'])


    def _on_each_record(self, next_row_number, next_record):
        """先読みで最初の１回を空振りさせるので、２件目から本処理です"""

        # 事前送り出し
        self._forward_cursor(next_record=next_record)

        if self._curr_record.no is None:
            if self._debug_write:
                # 最初のレコードは先読みのため、空回しします
                print(f"[{datetime.datetime.now()}] Pencil {self._curr_record.no} record  first record read later")


        else:
            # 変数名短縮
            ws = self._ws


            # ３行目～６行目
            # --------------
            # データは３行目から、１かたまり３行を使って描画する
            HEADER_HEIGHT = 3
            RECORD_HEIGHT = 3
            curr_row_number = next_row_number - 1
            row1_th = curr_row_number * RECORD_HEIGHT + HEADER_HEIGHT
            row2_th = curr_row_number * RECORD_HEIGHT + HEADER_HEIGHT + 1
            row3_th = curr_row_number * RECORD_HEIGHT + HEADER_HEIGHT + 2
            three_row_numbers = [row1_th, row2_th, row3_th]

            # 行の高さ設定
            # height の単位はポイント。昔のアメリカ人が椅子に座ってディスプレイを見たとき 1/72 インチに見える大きさが 1ポイント らしいが、そんなんワカラン。目視確認してほしい
            height = self._settings_obj.dictionary['row_height_of_upper_side_of_node']
            if height is not None:
                ws.row_dimensions[row1_th].height = height
            
            height = self._settings_obj.dictionary['row_height_of_lower_side_of_node']
            if height is not None:
                ws.row_dimensions[row2_th].height = height
            
            height = self._settings_obj.dictionary['row_height_of_node_spacing']
            if height is not None:
                ws.row_dimensions[row3_th].height = height


            ws[f'A{row1_th}'].value = self._curr_record.no
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'A{row1_th}'], index=0)
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'A{row2_th}'], index=0)
            self._settings_obj.set_bgcolor_of_header_to(cell=ws[f'A{row3_th}'], index=0)

            # 根側のパディング
            # ----------------
            cells = [
                ws[f'B{row1_th}'],
                ws[f'B{row2_th}'],
                ws[f'B{row3_th}'],
            ]
            for cell in cells:
                self._settings_obj.set_bgcolor_of_tree_to(cell=cell)


            def draw_edge(depth_th, three_column_names, three_row_numbers):
                """
                Parameters
                ----------
                depth_th : int
                    第何層。根層は 0
                """

                cn1 = three_column_names[0]
                cn2 = three_column_names[1]
                cn3 = three_column_names[2]
                row1_th = three_row_numbers[0]
                row2_th = three_row_numbers[1]
                row3_th = three_row_numbers[2]


                # ツリー構造図の背景色
                # --------------------
                cells = [
                    ws[f'{cn1}{row1_th}'],
                    ws[f'{cn1}{row2_th}'],
                    ws[f'{cn1}{row3_th}'],
                    ws[f'{cn2}{row1_th}'],
                    ws[f'{cn2}{row2_th}'],
                    ws[f'{cn2}{row3_th}'],
                    ws[f'{cn3}{row1_th}'],
                    ws[f'{cn3}{row2_th}'],
                    ws[f'{cn3}{row3_th}'],
                ]
                for cell in cells:
                    self._settings_obj.set_bgcolor_of_tree_to(cell=cell)

                # ↑ノードが無くても背景色は塗る必要がある


                nd = self._curr_record.node_at(depth_th=depth_th)

                if nd is None or pd.isnull(nd.text):
                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  Empty cell")
                    return


                # 自件と前件を比較して、根から自ノードまで、ノードテキストが等しいか？
                if LineByLineLeafModel.is_same_path_as_avobe(
                        curr_record=self._curr_record,
                        prev_record=self._prev_record,
                        depth_th=depth_th):

                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  │")
                    
                    # 垂直線
                    #
                    #   |    leftside_border
                    # ..+..  
                    #   |    leftside_border
                    #   |    leftside_border
                    #
                    self._settings_obj.set_leftside_border_to_vertical(cell=ws[f'{cn2}{row1_th}'])
                    self._settings_obj.set_leftside_border_to_vertical(cell=ws[f'{cn2}{row2_th}'])
                    self._settings_obj.set_leftside_border_to_vertical(cell=ws[f'{cn2}{row3_th}'])
                    return


                # 子ノードへの接続は４種類の線がある
                #
                # (1) ─字
                #   .    under_border
                # ...__  
                #   .    None
                #   .    None
                #
                # (2) ┬字
                #   .    under_border
                # ..+__  
                #   |    leftside_border
                #   |    leftside_border
                #
                # (3) ├字
                #   |    l_letter_border
                # ..+__  
                #   |    leftside_border
                #   |    leftside_border
                #
                # (4) └字
                #   |    l_letter_border
                # ..+__  
                #   .    None
                #   .    None
                #
                kind = LineByLineLeafModel.get_kind_of_edge(
                        prev_record=self._prev_record,
                        curr_record=self._curr_record,
                        next_record=self._next_record,
                        depth_th=depth_th)

                if kind == '─字':
                    self._settings_obj.set_border_to_parent_horizontal(cell=ws[f'{cn1}{row1_th}'])
                    self._settings_obj.set_under_border_to_child_horizontal(cell=ws[f'{cn2}{row1_th}'])
                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  ─ {nd.edge_text}")
                
                elif kind == '┬字':
                    self._settings_obj.set_border_to_parent_downward(cell=ws[f'{cn1}{row1_th}'])
                    self._settings_obj.set_under_border_to_child_downward(cell=ws[f'{cn2}{row1_th}'])
                    self._settings_obj.set_leftside_border_to_child_downward(cell=ws[f'{cn2}{row2_th}'])
                    self._settings_obj.set_leftside_border_to_child_downward(cell=ws[f'{cn2}{row3_th}'])
                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  ┬ {nd.edge_text}")

                elif kind == '├字':
                    self._settings_obj.set_l_letter_border_to_child_rightward(cell=ws[f'{cn2}{row1_th}'])
                    self._settings_obj.set_leftside_border_to_child_rightward(cell=ws[f'{cn2}{row2_th}'])
                    self._settings_obj.set_leftside_border_to_child_rightward(cell=ws[f'{cn2}{row3_th}'])
                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  ├ {nd.edge_text}")

                elif kind == '└字':
                    self._settings_obj.set_l_letter_border_to_child_upward(cell=ws[f'{cn2}{row1_th}'])
                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Edge) {self._curr_record.no} record > {nth(depth_th)} layer  └ {nd.edge_text}")
                
                else:
                    raise ValueError(f"{kind=}")
                

                # ２列目：エッジ・テキスト
                ws[f'{cn2}{row1_th}'].value = nd.edge_text


            def draw_node(depth_th, three_column_names, three_row_numbers):
                """節を描きます

                Parameters
                ----------
                node : NodeInRecord
                    節
                depth_th : int
                    第何層。根層は 0
                """

                cn3 = three_column_names[2]
                row1_th = three_row_numbers[0]
                row2_th = three_row_numbers[1]
                row3_th = three_row_numbers[2]

                nd = self._curr_record.node_at(depth_th=depth_th)

                if nd is None or pd.isnull(nd.text) or LineByLineLeafModel.is_same_path_as_avobe(
                        curr_record=self._curr_record,
                        prev_record=self._prev_record,
                        depth_th=depth_th):

                    if self._debug_write:
                        print(f"[{datetime.datetime.now()}] Pencil(Node) {self._curr_record.no} record > {nth(depth_th)} layer  Empty cell")

                    # ツリー構造図の背景色
                    # --------------------
                    cells = [
                        ws[f'{cn3}{row1_th}'],
                        ws[f'{cn3}{row2_th}'],
                        ws[f'{cn3}{row3_th}'],
                    ]
                    for cell in cells:
                        self._settings_obj.set_bgcolor_of_tree_to(cell=cell)

                    return



                if self._debug_write:
                    print(f"[{datetime.datetime.now()}] Pencil(Node) {self._curr_record.no} record > {nth(depth_th)} layer  □ {nd.text}")
                
                ws[f'{cn3}{row1_th}'].value = nd.text
                self._settings_obj.set_alignment_of_node_to(cell=ws[f'{cn3}{row1_th}'])
                self._settings_obj.set_bgcolor_of_node_to(cell=ws[f'{cn3}{row1_th}'])
                self._settings_obj.set_border_of_upside_node(cell=ws[f'{cn3}{row1_th}'])

                self._settings_obj.set_bgcolor_of_node_to(cell=ws[f'{cn3}{row2_th}'])
                self._settings_obj.set_border_of_downside_node(cell=ws[f'{cn3}{row2_th}'])

                self._settings_obj.set_bgcolor_of_tree_to(cell=ws[f'{cn3}{row3_th}'])      # ツリー構造図の背景色

                # セル結合していいなら
                if not self._settings_obj.dictionary['do_not_merge_cells']:
                    #print(f"セル結合 {cn3}{row1_th}:{cn3}{row2_th}")
                    ws.merge_cells(f'{cn3}{row1_th}:{cn3}{row2_th}')

                    # FIXME 文字も数も区別なく左に揃えていいものか？
                    #self._settings_obj.set_alignment_left_center(cell=ws[f'{cn3}{row1_th}'])
                    self._settings_obj.set_alignment_center_center(cell=ws[f'{cn3}{row1_th}'])


            # 第０層
            # ------
            depth_th = 0
            column_letter = xl.utils.get_column_letter(3)   # 'C'
            if depth_th < self._table.analyzer.end_th_of_node:
                draw_node(depth_th=depth_th, three_column_names=[None, None, column_letter], three_row_numbers=three_row_numbers)


            # 第１～最終層
            # ------------
            for depth_th in range(1, self._table.analyzer.end_th_of_node):
                head_column_th = depth_th * StyleControl.ONE_NODE_COLUMNS + 1

                if depth_th < self._table.analyzer.end_th_of_node:
                    # 第1層は 'D', 'E', 'F'、以降、後ろにずれていく
                    column_letter_list = [
                        xl.utils.get_column_letter(head_column_th),
                        xl.utils.get_column_letter(head_column_th + 1),
                        xl.utils.get_column_letter(head_column_th + 2),
                    ]
                    draw_edge(depth_th=depth_th, three_column_names=column_letter_list, three_row_numbers=three_row_numbers)
                    draw_node(depth_th=depth_th, three_column_names=column_letter_list, three_row_numbers=three_row_numbers)


            column_name_of_last_node = self._table.analyzer.get_column_name_of_last_node()


            # 葉側のパディング
            # ----------------
            column_th = StyleControl.get_target_column_th(source_table=self._table, column_name=column_name_of_last_node) + 1
            column_letter = xl.utils.get_column_letter(column_th)
            cells = [
                ws[f'{column_letter}{row1_th}'],
                ws[f'{column_letter}{row2_th}'],
                ws[f'{column_letter}{row3_th}'],
            ]
            for cell in cells:
                self._settings_obj.set_bgcolor_of_tree_to(cell=cell)      # ツリー構造図の背景色



            # 余り列
            # ------
            # 最終層以降の列
            is_remaining = False
            for column_name in self._table.df.columns:
                if column_name == column_name_of_last_node:
                    is_remaining = True
                    continue

                elif is_remaining:
                    # TODO キャッシュを作れば高速化できそう
                    target_column_th = StyleControl.get_target_column_th(source_table=self._table, column_name=column_name)
                    column_letter = xl.utils.get_column_letter(target_column_th)

                    #print(f'{row1_th=}  {column_name=}')
                    ws[f'{column_letter}{row1_th}'].value = self._table.df.at[curr_row_number + 1, column_name]

                    # 罫線
                    self._settings_obj.set_upper_border_of_remaining_cell(cell=ws[f'{column_letter}{row1_th}'])
                    self._settings_obj.set_middle_border_of_remaining_cell(cell=ws[f'{column_letter}{row2_th}'])
                    self._settings_obj.set_lower_border_of_remaining_cell(cell=ws[f'{column_letter}{row3_th}'])

                    # ツリー構造図の背景色
                    cells = [
                        ws[f'{column_letter}{row1_th}'],
                        ws[f'{column_letter}{row2_th}'],
                        ws[f'{column_letter}{row3_th}'],
                    ]
                    for cell in cells:
                       self._settings_obj.set_bgcolor_of_tree_to(cell=cell)

                    # セル結合していいなら
                    if not self._settings_obj.dictionary['do_not_merge_cells']:
                        ws.merge_cells(f'{column_letter}{row1_th}:{column_letter}{row3_th}')

                        # FIXME 文字も数も区別なく左に揃えていいものか？
                        #self._settings_obj.set_alignment_left_center(cell=ws[f'{column_letter}{row1_th}'])
                        self._settings_obj.set_alignment_center_center(cell=ws[f'{column_letter}{row1_th}'])


class TreeEraser():
    """要らない罫線を消す"""


    def __init__(self, table, ws, settings_obj, debug_write=False):
        """初期化
        
        Parameters
        ----------
        table : Table
            ツリーテーブル
        ws : openpyxl.Worksheet
            ワークシート
        settings_obj : Settings
            各種設定
        debug_write : bool
            デバッグライト
            DEBUG_TIPS: デバッグライトをオンにして、コンソールにログを表示すると不具合を調査しやすくなります
        """
        self._table = table
        self._ws = ws
        self._settings_obj = settings_obj
        self._debug_write = debug_write


    def render(self):
        """描画"""

        # 指定の列の左側の垂直の罫線を見ていく
        column_th = 5
        for node_th in range(1, self._table.analyzer.end_th_of_node):
            self._erase_unnecessary_border_by_column(column_letter=xl.utils.get_column_letter(column_th))
            column_th += 3


    def _erase_unnecessary_border_by_column(self, column_letter):
        """不要な境界線を消す"""

        # 変数名の短縮
        ws = self._ws


        # 最後に見つけた、セルの左辺に罫線がなく、下辺に太い罫線がある行をリセット
        row_th_of_prev_last_underline = -1
        row_th_of_last_underline = -1


        # 第3行から
        row_th = 3
        while row_th <= ws.max_row: # 最終行まで全部見る

            # 前行のセルには、左辺と可変に太い罫線があったか？
            prerow_l_letter = False

            while True: # 仕切り直しの１セット
                shall_break = False

                currow_l_letter = False

                # 罫線を確認
                #
                #   .
                # ..+--  下向きの罫線が最後に出た箇所を調べる
                #   |
                #
                border = ws[f'{column_letter}{row_th}'].border
                if border is not None:
                    # セルの左辺に太い罫線が引かれており...
                    if border.left is not None and border.left.style == 'thick':
                        # セルの下辺にも太い罫線が引かれていれば、'└' 字か '├' 字のどちらかだ
                        if border.bottom is not None and border.bottom.style == 'thick':
                            row_th_of_prev_last_underline = row_th_of_last_underline
                            row_th_of_last_underline = row_th
                            currow_l_letter = True
                            if self._debug_write:
                                # 左側と下側に罫線。 '└' 字か '├' 字のどちらかだ。アンダーラインが第何行か覚えておく
                                print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} Border on the left and bottom. '└' or '├'. Memory underline ({row_th_of_last_underline} row) (Preview {row_th_of_prev_last_underline} row)")

                        # 左辺に罫線。次行へ読み進めていく
                        else:
                            if self._debug_write:
                                print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} Border on the left")

                    # セルの左辺に太い罫線が引かれていない
                    else:
                        # "└"字。［ラスト・シブリング］なので、最後に見つけた左辺に罫線のないアンダーラインのことは忘れて仕切り直し
                        if prerow_l_letter:
                            row_th_of_prev_last_underline = -1
                            row_th_of_last_underline = -1
                            shall_break = True
                            if self._debug_write:
                                print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} Last sibling. Forget underline row")

                        # セルの下辺に太い罫線が引かれていたら、つながっていない垂線だ。それが第何行か覚えておいて仕切り直す
                        elif border.bottom is not None and border.bottom.style == 'thick':
                            row_th_of_prev_last_underline = row_th_of_last_underline
                            row_th_of_last_underline = row_th
                            shall_break = True
                            if self._debug_write:
                                print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} Memory underline ({row_th_of_last_underline} row) (Preview {row_th_of_prev_last_underline} row)")

                        # セルの左辺にも、下辺にも、太い罫線が引かれていなければ、罫線は尻切れトンボになっている。仕切り直し
                        else:
                            row_th_of_prev_last_underline = row_th_of_last_underline
                            row_th_of_last_underline = row_th
                            shall_break = True
                            if self._debug_write:
                                print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} No border on the left and bottom. no connect line. Memory ({row_th_of_last_underline} row) (Preview {row_th_of_prev_last_underline} row)")


                row_th += 1

                prerow_l_letter = currow_l_letter

                if shall_break:
                    break


            # 消しゴムを掛ける
            start_row_to_erase = row_th_of_prev_last_underline + 1
            end_row_to_erase = row_th_of_last_underline

            if row_th_of_last_underline != -1 and 0 < start_row_to_erase and start_row_to_erase < end_row_to_erase:

                if self._debug_write:
                    print(f"[{datetime.datetime.now()}] Eraser {column_letter}_ Erase {start_row_to_erase} to {end_row_to_erase - 1} row...")

                for row_th_to_erase in range(start_row_to_erase, end_row_to_erase):
                    # 消すか、見え消しにするか切り替えられるようにしておく
                    self._settings_obj.set_striked_border(cell=ws[f'{column_letter}{row_th_to_erase}'])

        if self._debug_write:
            print(f"[{datetime.datetime.now()}] Eraser {column_letter}{row_th} finished (EOL {ws.max_row})")



import datetime
import pandas as pd

from ...library import INDENT
from .table_formatting import ColumnsSorting, InputCompletion
from .source_csv_table_analyzer import SourceCsvTableAnalyzer


############
# MARK: Node
############
class NodeInRecord():
    """ノード（節）
    ノードの形に合わせて改造してください"""

    def __init__(self, edge_text, text):
        """初期化
        
        Parameters
        ----------
        edge_text : str
            辺のテキスト
        text : str
            節のテキスト
        """
        self._edge_text = edge_text
        self._text = text


    @property
    def edge_text(self):
        return self._edge_text


    @property
    def text(self):
        return self._text


    def _pack_key(self):
        return (self._edge_text, self._text)


    def _stringify_dump(self, indent):
        succ_indent = indent + INDENT
        return f"""\
{indent}NodeInRecord
{indent}------------
{succ_indent}{self._edge_text=}
{succ_indent}{self._text=}
"""


##############
# MARK: Record
##############
class Record():


    def __init__(self, no, root_to_leaf_pathway):
        """初期化
        
        Parameters
        ----------
        no : int
            1から始まる連番。数詞は件
        root_to_leaf_pathway : list<NodeInRecord>
            根から葉まで並んだノードのリスト。
            第０層は根
        """
        self._no = no
        self._root_to_leaf_pathway = root_to_leaf_pathway


    @staticmethod
    def new_empty(specified_end_th_of_node):
        return Record(
                no=None,
                root_to_leaf_pathway=[])


    @property
    def no(self):
        return self._no


    @property
    def len_of_path_from_root_to_leaf(self):
        return len(self._root_to_leaf_pathway)


    def node_at(self, depth_th):
        """
        Parameters
        ----------
        round_th : int
            th は forth や fifth の th。
            例：根なら０を指定してください。
            例：第１層なら 1 を指定してください
        
        Returns
        -------
        node : NodeInRecord
            ノード、なければナン
        """

        # NOTE -1 を指定すると最後尾の要素になるが、固定長配列の最後尾の要素が、思っているような最後尾の要素とは限らない。うまくいかない
        if depth_th < 0:
            raise ValueError(f'depth_th に負数を設定しないでください。意図した動作はしません {depth_th=}')

        if depth_th < len(self._root_to_leaf_pathway):
            return self._root_to_leaf_pathway[depth_th]

        return None


    def update(self, no=None, root_to_leaf_pathway=None):
        """no inplace
        何も更新しなければシャロー・コピーを返します"""

        def new_or_default(new, default):
            if new is None:
                return default
            return new

        return Record(
                no=new_or_default(no, self._no),
                root_to_leaf_pathway=new_or_default(root_to_leaf_pathway, self._root_to_leaf_pathway))


    def _stringify_dump(self, indent):
        succ_indent = indent + INDENT

        blocks = []
        for node in self._root_to_leaf_pathway:
            blocks.append(node._stringify_dump(succ_indent))

        return f"""\
{indent}Record
{indent}------
{succ_indent}{self._no=}
{'\n'.join(blocks)}
"""


    def for_each_node_in_path(self, set_node):
        for depth, node in enumerate(self._root_to_leaf_pathway):
            set_node(depth, node)


    def get_th_of_leaf_entry(self):
        """葉要素の層番号を取得。
        th は forth や fifth の th。
        葉要素は、次の層がない要素"""

        for depth_th in range(0, len(self._root_to_leaf_pathway)):
            nd = self._root_to_leaf_pathway[depth_th]
            if nd is None or nd.text is None:
                return depth_th

        return len(self._root_to_leaf_pathway)


#############
# MARK: Table
#############
class Table():
    """樹形図データのテーブル"""


    # 列が可変長
    _dtype = {}

    @classmethod
    def create_dtype(clazz, specified_end_th_of_edge, specified_end_th_of_node):
        """dtypeの辞書を作成します

        Parameters
        ----------
        specified_end_th_of_edge : int
            エッジ数。空欄の根を含むとみなして数える
        specified_end_th_of_node : int
            ノード数。根を含む
        """

        # no はインデックスなので含めない
        clazz._dtype = {}

        # ノードだけ根を含む
        clazz._dtype['node0'] = 'object'    # string 型は無いので object 型にする

        for edge_th in range(1, specified_end_th_of_edge):
            clazz._dtype[f'edge{edge_th}'] = 'object'

        for node_th in range(1, specified_end_th_of_node):
            clazz._dtype[f'node{node_th}'] = 'object'

        return clazz._dtype


    @staticmethod
    def create_column_name_list(specified_end_th_of_node, include_index):
        column_name_list = []

        if include_index:
            column_name_list.append('no')

        # 根ノードは必ず追加
        column_name_list.append('node0')

        for node_th in range(1, specified_end_th_of_node):
            column_name_list.append(f'edge{node_th}')
            column_name_list.append(f'node{node_th}')

        return column_name_list


    def __init__(self, df, analyzer):
        self._df = df
        self._analyzer = analyzer


    @classmethod
    def new_empty_table(clazz, specified_end_th_of_edge, specified_end_th_of_node):
        column_name_list = Table.create_column_name_list(
                specified_end_th_of_node=specified_end_th_of_node,
                include_index=True) # 'no' は後でインデックスに変換

        df = pd.DataFrame(
                columns=column_name_list)
        
        clazz.setup_data_frame(df=df, specified_end_th_of_edge=specified_end_th_of_edge, specified_end_th_of_node=specified_end_th_of_node, shall_set_index=True)

        # 元テーブルの分析器
        analyzer = SourceCsvTableAnalyzer.instantiate(df=df, end_th_of_edge=0, end_th_of_node=0)


        return Table(df=df, analyzer=analyzer)


    @classmethod
    def from_csv(clazz, file_path):
        """ファイル読込

        Parameters
        ----------
        file_path : str
            CSVファイルパス
        
        Returns
        -------
        table : Table
            テーブル、またはナン
        file_read_result : FileReadResult
            ファイル読込結果
        """

        # 'no' 列が有るかどうか分からないので、読み込み時に index_col は指定できません
        df = pd.read_csv(file_path, encoding="utf8") #index_col=['no']

        # no 列が含まれていないなら、１から始まる自動連番を追加します
        if 'no' not in df.columns:
            df['no'] = range(1, len(df.index) + 1)

        # 列名をソートしたい。no,node0,edge1,node1,edge2,node2,remaining_a,remaining_b,... のような感じに
        #print(f"列ソート前 {df.columns.values=}")
        columns_sorting = ColumnsSorting()
        df = columns_sorting.execute(df)
        #print(f"列ソート後 {df.columns.values=}")



        # テーブルに追加の設定
        clazz.setup_data_frame(df=df, specified_end_th_of_edge=columns_sorting.end_th_of_edge, specified_end_th_of_node=columns_sorting.end_th_of_node,
            shall_set_index=True) # 'no' 列をインデックスに指定します


        # 整形
        InputCompletion.fill_directory(df=df, end_th_of_edge=columns_sorting.end_th_of_edge, end_th_of_node=columns_sorting.end_th_of_node)

        return Table(
            df=df,
            # 元テーブルの分析器
            analyzer=SourceCsvTableAnalyzer.instantiate(df=df, end_th_of_edge=columns_sorting.end_th_of_edge, end_th_of_node=columns_sorting.end_th_of_node))


    @property
    def df(self):
        return self._df


    @property
    def analyzer(self):
        return self._analyzer


    @classmethod
    def setup_data_frame(clazz, df, specified_end_th_of_edge, specified_end_th_of_node, shall_set_index):
        """データフレームの設定"""

        if shall_set_index:
            # インデックスの設定
            df.set_index('no',
                    inplace=True)   # NOTE インデックスを指定したデータフレームを戻り値として返すのではなく、このインスタンス自身を更新します

        # データ型の設定
        dtype = clazz.create_dtype(specified_end_th_of_edge=specified_end_th_of_edge, specified_end_th_of_node=specified_end_th_of_node)
        #print(f"[{datetime.datetime.now()}] setup_data_frame {dtype=}")
        df.astype(dtype)


    def upsert_record(self, welcome_record):
        """該当レコードが無ければ新規作成、あれば更新

        Parameters
        ----------
        welcome_record : GameTreeRecord
            レコード

        Returns
        -------
        shall_record_change : bool
            レコードの新規追加、または更新があれば真。変更が無ければ偽
        """

        # インデックス
        # -----------
        # index : any
        #   インデックス。整数なら numpy.int64 だったり、複数インデックスなら tuple だったり、型は変わる。
        #   <class 'numpy.int64'> は int型ではないが、pandas では int型と同じように使えるようだ
        index = welcome_record.no

        # データ変更判定
        # -------------
        is_new_index = index not in self._df.index

        # インデックスが既存でないなら
        if is_new_index:
            shall_record_change = True

        else:
            # 更新の有無判定
            shall_record_change = True
            # no はインデックスなので含めない

            # 根は必ず含める
            if self._df['node0'][index] != welcome_record.node_at(0).text:
                shall_record_change = False
            
            for node_th in range(1, self._analyzer.end_th_of_node):
                if self._df[f'node{node_th}'][index] != welcome_record.node_at(node_th).text:
                    shall_record_change = False
                    break


        # 行の挿入または更新
        if shall_record_change:

            # no はインデックスなので含めない
            dictionary = {}

            # 根は必ず含める
            dictionary['node0'] = welcome_record.node_at(0).text

            for node_th in range(1, self.end_th_of_node):
                dictionary[f'edge{node_th}'] = welcome_record.node_at(node_th).edge_text
                dictionary[f'node{node_th}'] = welcome_record.node_at(node_th).text

            self._df.loc[index] = dictionary


        if is_new_index:
            # NOTE ソートをしておかないと、インデックスのパフォーマンスが機能しない
            self._df.sort_index(
                    inplace=True)   # NOTE ソートを指定したデータフレームを戻り値として返すのではなく、このインスタンス自身をソートします


        return shall_record_change


    def to_csv(self, file_path):
        """ファイル書き出し
        
        Parameters
        ----------
        file_path : str
            CSVファイルパス
        """

        column_name_list = Table.create_column_name_list(
                specified_end_th_of_node=self.end_th_of_node,
                include_index=False) # no はインデックスなので含めない

        self._df.to_csv(
                csv_file_path,
                columns=column_name_list)


    def for_each(self, on_each):
        """
        Parameters
        ----------
        on_each : func
            Record 型引数を受け取る関数
        """

        df = self._df

        root_to_leaf_pathway = [None] * self._analyzer.end_th_of_node

        for row_number in range(0, len(df)):
            # no はインデックス
            no = df.index[row_number]

            root_to_leaf_pathway = []

            # 根
            root_to_leaf_pathway.append(NodeInRecord(edge_text=None, text=df.at[no, f'node0']))

            # 中間～葉ノード
            for node_th in range(1, self._analyzer.end_th_of_node):

                # ノードのテキスト
                node_text = df.at[no, f'node{node_th}']

                # ノードのテキストが未設定なら無視
                if pd.isnull(node_text):
                    continue

                # エッジはオプション
                if node_th < self._analyzer.end_th_of_edge:
                    edge_text = df.at[no, f'edge{node_th}']
                else:
                    edge_text = None

                root_to_leaf_pathway.append(NodeInRecord(edge_text=edge_text, text=node_text))


            # レコード作成
            record = Record(
                    no=no,
                    root_to_leaf_pathway=root_to_leaf_pathway)

            on_each(row_number, record)



#
# python example.py all
#
# 例を実行しよう
#

import traceback
import datetime
import sys


########################################
# コマンドから実行時
########################################
if __name__ == '__main__':
    """コマンドから実行時"""

    try:
        args = sys.argv

        if 1 < len(args):
            if args[1] == 'all':
                from examples.o1o0 import *
                from examples.o1o1o0 import *
                from examples.o2o0 import *
                from examples.o3o0 import *
                from examples.o4o0_model import *

            elif args[1] == 'o1o0':
                from examples.o1o0 import *

            elif args[1] == 'o1o1o0':
                from examples.o1o1o0 import *

            elif args[1] == 'o2o0':
                from examples.o2o0 import *

            elif args[1] == 'o3o0':
                from examples.o3o0 import *

            elif args[1] == 'o4o0':
                from examples.o4o0_model import *

            else:
                raise ValueError(f'unsupported {args[1]=}')
        
        else:
            raise ValueError(f'unsupported {len(args)=}')


    except Exception as err:
        print(f"""\
[{datetime.datetime.now()}] おお、残念！　例外が投げられてしまった！
{type(err)=}  {err=}

以下はスタックトレース表示じゃ。
{traceback.format_exc()}
""")



#
# python test.py all
#
# エクセルで樹形図を描こう
#

import traceback
import datetime
import sys

from tests.e_o1o0 import execute as execute_e_o1o0
from tests.e_o2o0 import execute as execute_e_o2o0
from tests.e_o3o0 import execute as execute_e_o3o0
from tests.e_o4o0 import execute as execute_e_o4o0

from tests.manual import execute as execute_manual
from tests.t_o1o0 import execute as execute_t_o1o0
from tests.t_o1o1o0_tree_model import execute as execute_t_o1o1o0
from tests.t_o2o0 import execute as execute_t_o2o0
from tests.t_o3o1o0 import execute as execute_t_o3o1o0
from tests.t_o3o2o0_do_not_merge_cells import execute as execute_t_o3o2o0
from tests.t_o4o0 import execute as execute_t_o4o0
from tests.t_o5o0 import execute as execute_t_o5o0
from tests.t_o6o0 import execute as execute_t_o6o0
from tests.t_o6o1o0_tree_model import execute as execute_t_o6o1o0
from tests.t_o7o0 import execute as execute_t_o7o0


########################################
# コマンドから実行時
########################################
if __name__ == '__main__':
    """コマンドから実行時"""

    try:
        args = sys.argv

        if 1 < len(args):

            if args[1] == 'all':
                execute_e_o1o0()
                execute_e_o2o0()
                execute_e_o3o0()
                execute_e_o4o0()

                execute_t_o1o0()
                execute_t_o1o1o0()
                execute_t_o2o0()
                execute_t_o3o1o0()
                execute_t_o3o2o0()
                execute_t_o4o0()
                execute_t_o5o0()
                execute_t_o6o0()
                execute_t_o6o1o0()
                execute_t_o7o0()

            elif args[1] == 'e_o1o0':
                execute_e_o1o0()

            elif args[1] == 'e_o2o0':
                execute_e_o2o0()

            elif args[1] == 'e_o3o0':
                execute_e_o3o0()

            elif args[1] == 'e_o4o0':
                execute_e_o4o0()

            elif args[1] == 't_o1o0':
                execute_t_o1o0()

            elif args[1] == 't_o1o1o0':
                execute_t_o1o1o0()

            elif args[1] == 't_o2o0':
                execute_t_o2o0()

            elif args[1] == 't_o3o1o0':
                execute_t_o3o1o0()

            elif args[1] == 't_o3o2o0':
                execute_t_o3o2o0()

            elif args[1] == 't_o4o0':
                execute_t_o4o0()

            elif args[1] == 't_o5o0':
                execute_t_o5o0()

            elif args[1] == 't_o6o0':
                execute_t_o6o0()

            elif args[1] == 't_o6o1o0':
                execute_t_o6o1o0()

            elif args[1] == 't_o7o0':
                execute_t_o7o0()

            else:
                raise ValueError(f'unsupported {args[1]=}')
        
        else:
            execute_manual()


    except Exception as err:
        print(f"""\
[{datetime.datetime.now()}] おお、残念！　例外が投げられてしまった！
{type(err)=}  {err=}

以下はスタックトレース表示じゃ。
{traceback.format_exc()}
""")
